use {
    super::B64Alphabet,
    std::{
        fmt,
        io::{self, ErrorKind, Read, Write},
        num::TryFromIntError,
    },
};

#[derive(Debug)]
/// Errors which might possibly occur while decoding base64 encoded data
pub enum Error {
    Io(io::Error),
    IllegalChar(char),
    MissingPadding,
    IntError(TryFromIntError),
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Io(e) => write!(f, "{e}"),
            Self::IllegalChar(c) => write!(f, "Illegal Character ({c})"),
            Self::MissingPadding => write!(f, "Missing Padding"),
            Self::IntError(e) => write!(f, "Int Error: {e}"),
        }
    }
}

impl From<io::Error> for Error {
    fn from(value: io::Error) -> Self {
        Self::Io(value)
    }
}

impl From<char> for Error {
    fn from(value: char) -> Self {
        Self::IllegalChar(value)
    }
}

impl From<TryFromIntError> for Error {
    fn from(value: TryFromIntError) -> Self {
        Self::IntError(value)
    }
}

impl std::error::Error for Error {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Self::Io(e) => Some(e),
            Self::IntError(e) => Some(e),
            _ => None,
        }
    }
}

pub struct Decoder<R: Read, W: Write> {
    reader: R,
    writer: W,
    alphabet: B64Alphabet,
}

impl<R: Read, W: Write> Decoder<R, W> {
    pub fn new(reader: R, writer: W, alphabet: Option<B64Alphabet>) -> Self {
        Self {
            reader,
            writer,
            alphabet: alphabet.unwrap_or_default(),
        }
    }

    /// Decodes the bytes provided by `self.reader` and writes the resulting
    /// bytes into `self.writer`.
    /// # Errors
    /// - io failure will return `Error::Io`
    /// - a character not in the alphabet, or encountering any character after
    ///   the first occurance of the padding character, will return `Error::IllegalChar`
    /// - if the reader comes up short of the final four byte block will return
    ///   `Error::MissingPadding`
    /// - `intError` should never be returned. If this error is recieved the code
    ///   is somehow broken. Please file a bug.
    pub fn decode(&mut self) -> Result<(), Error> {
        loop {
            let mut in_buf = [0_u8; 4];
            let mut out_buf = [0_u8; 3];
            let mut num = 0_u32;
            let mut n_bytes = 0;
            loop {
                n_bytes += match self.reader.read(&mut in_buf) {
                    Ok(n) => n,
                    Err(e) if e.kind() == ErrorKind::Interrupted => continue,
                    Err(e) => return Err(e.into()),
                };
                break;
            }
            match n_bytes {
                0 => break,
                4 => {}
                _ => return Err(Error::MissingPadding),
            }
            let mut bytes = 0;
            for (i, &c) in in_buf.iter().enumerate() {
                let c = c.into();
                num <<= 6;
                if c == self.alphabet.pad() {
                    continue;
                }
                if i != bytes {
                    return Err(Error::IllegalChar(c));
                }
                let Some(idx) = self.alphabet.get(c) else {
                    return Err(Error::IllegalChar(c));
                };
                num |= u32::try_from(idx)?;
                bytes += 1;
            }
            let olen = bytes * 6 / 8;
            for i in (0..3).rev() {
                out_buf[i] = (num & 0xff) as u8;
                num >>= 8;
            }
            self.writer.write_all(&out_buf[0..olen])?;
        }
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn decode() {
        let mut decoder = Decoder::new("SGVsbG8sIFdvcmxk".as_bytes(), vec![], None);
        decoder.decode().unwrap();
        assert_eq!(String::from_utf8(decoder.writer).unwrap(), "Hello, World");
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQ==".as_bytes(), vec![], None);
        decoder.decode().unwrap();
        assert_eq!(String::from_utf8(decoder.writer).unwrap(), "Hello, World!");
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQo=".as_bytes(), vec![], None);
        decoder.decode().unwrap();
        assert_eq!(
            String::from_utf8(decoder.writer).unwrap(),
            "Hello, World!\n"
        );
    }
}
